from ctypes import POINTER, c_double

import numpy as np


def test_parallel_edgecases(NYC_llvm, NYC_lgbm):
    # single row, multiple threads
    data = np.array(1 * [NYC_lgbm.num_feature() * [1.0]], dtype=np.float64)
    np.testing.assert_almost_equal(
        NYC_llvm.predict(data, n_jobs=4), NYC_lgbm.predict(data), decimal=14
    )

    # last thread has only one prediction (batchsize is ceil(19/7)=3)
    data = np.array(19 * [NYC_lgbm.num_feature() * [1.0]], dtype=np.float64)
    np.testing.assert_almost_equal(
        NYC_llvm.predict(data, n_jobs=7), NYC_lgbm.predict(data), decimal=14
    )


def test_parallel_iteration(NYC_llvm, NYC_lgbm):
    data = np.array(4 * [NYC_lgbm.num_feature() * [1.0]], dtype=np.float64)
    data_flat = np.array(data.reshape(data.size), dtype=np.float64)
    np.testing.assert_almost_equal(
        NYC_llvm.predict(data, n_jobs=4), NYC_lgbm.predict(data), decimal=14
    )

    ptr_data = data_flat.ctypes.data_as(POINTER(c_double))
    preds = np.zeros(4, dtype=np.float64)
    ptr_preds = preds.ctypes.data_as(POINTER(c_double))

    NYC_llvm._c_entry_func(ptr_data, ptr_preds, 2, 4)
    preds_l = list(preds)
    assert preds_l[0] == 0.0 and preds_l[1] == 0.0
    assert preds_l[2] != 0.0 and preds_l[3] != 0.0
    NYC_llvm._c_entry_func(ptr_data, ptr_preds, 0, 2)
    preds_l = list(preds)
    assert preds_l[0] != 0.0 and preds_l[1] != 0.0



import pytest
from lightgbm import Booster

from lleaves import Model


@pytest.fixture(scope="session")
def NYC_llvm():
    llvm_model = Model(model_file="tests/models/NYC_taxi/model.txt")
    llvm_model.compile()
    return llvm_model


@pytest.fixture(scope="session")
def NYC_lgbm():
    return Booster(model_file="tests/models/NYC_taxi/model.txt")


@pytest.fixture(scope="session")
def mtpl2_llvm():
    llvm_model = Model(model_file="tests/models/mtpl2/model.txt")
    llvm_model.compile()
    return llvm_model



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from lightgbm import Booster

from lleaves import Model
from lleaves.data_processing import (
    data_to_ndarray,
    extract_model_global_features,
    extract_pandas_traintime_categories,
    ndarray_to_ptr,
)


def test_parsing_pandas(tmp_path):
    mod_model_file = tmp_path / "mod_model.txt"
    model_file = Path("tests/models/pure_categorical/model.txt")
    with open(model_file, "r") as file:
        lines = file.readlines()
    assert lines[-1].startswith("pandas_categorical")
    lines[
        -1
    ] = 'pandas_categorical:[["a", "b", "c"], ["b", "c", "d"], ["w", "x", "y", "z"]]'

    with open(mod_model_file, "x") as file:
        file.writelines(lines)

    pandas_categorical = extract_pandas_traintime_categories(model_file)
    assert pandas_categorical is None
    pandas_categorical = extract_pandas_traintime_categories(mod_model_file)
    assert pandas_categorical == [
        ["a", "b", "c"],
        ["b", "c", "d"],
        ["w", "x", "y", "z"],
    ]


def test_parsing_pandas_broken_file(tmp_path):
    mod_model_file = tmp_path / "mod_model.txt"
    lines = 100 * ["onelineonly"]
    with open(mod_model_file, "x") as file:
        file.writelines(lines)

    # terminates and raises on garbage file
    with pytest.raises(ValueError):
        _ = extract_pandas_traintime_categories(mod_model_file)


def test_n_args_extract(tmp_path):
    mod_model_file = tmp_path / "mod_model.txt"
    model_file = Path("tests/models/mixed_categorical/model.txt")
    with open(model_file, "r") as file:
        lines = file.readlines()

    with open(mod_model_file, "x") as file:
        file.writelines(
            (line for line in lines if not line.startswith("max_feature_idx"))
        )

    res = extract_model_global_features(model_file)
    assert res["n_class"] == 1
    assert res["n_feature"] == 5
    with pytest.raises(ValueError):
        extract_model_global_features(mod_model_file)


def test_no_data_modification():
    # the data shouldn't be modified during conversion
    data = [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]]
    for dtype in [np.float32, np.float64, np.int64]:
        orig = np.array(data, dtype=dtype)
        pred = np.array(data, dtype=dtype)

        ndarray_to_ptr(data_to_ndarray(pred))

        np.testing.assert_array_equal(pred, orig)
        assert pred.dtype == orig.dtype

    for dtype in [np.float32, np.float64, np.int64]:
        orig = pd.DataFrame(data).astype(dtype)
        pred = pd.DataFrame(data).astype(dtype)
        ndarray_to_ptr(data_to_ndarray(pred, []))
        pd.testing.assert_frame_equal(pred, orig)

    data = [["a", "b"], ["b", "a"]]
    orig = pd.DataFrame(data).astype("category")
    pred = pd.DataFrame(data).astype("category")
    ndarray_to_ptr(data_to_ndarray(pred, data))
    pd.testing.assert_frame_equal(pred, orig)


def test_sliced_arrays():
    # predictions should be correct when passed a sliced array
    llvm_model = Model(model_file="tests/models/single_tree/model.txt")
    llvm_model.compile()
    lgbm_model = Booster(model_file="tests/models/single_tree/model.txt")

    n_feature = lgbm_model.num_feature()
    data = np.array(list(range(-5 * n_feature, 5 * n_feature)), dtype=np.float64)
    data = data.reshape((5, 2 * n_feature))
    sliced = data[:, ::2]
    assert not sliced.flags.c_contiguous
    np.testing.assert_almost_equal(
        llvm_model.predict(sliced, n_jobs=4), lgbm_model.predict(sliced), decimal=13
    )
    return



import io
import os
from contextlib import redirect_stdout

import numpy as np
import pandas as pd
import pytest
from lightgbm import Booster

from benchmarks.benchmark import NYC_used_columns
from benchmarks.train_NYC_model import feature_enginering
from lleaves import Model


@pytest.fixture(scope="session")
def NYC_data():
    df = pd.read_parquet(
        "benchmarks/data/yellow_tripdata_2016-01.parquet", columns=NYC_used_columns
    )
    return feature_enginering().fit_transform(df).astype(np.float64)


@pytest.mark.parametrize("blocksize", [1, 34, 100])
def test_cache_blocksize(blocksize, NYC_data):
    llvm_model = Model(model_file="tests/models/NYC_taxi/model.txt")
    lgbm_model = Booster(model_file="tests/models/NYC_taxi/model.txt")

    os.environ["LLEAVES_PRINT_UNOPTIMIZED_IR"] = "1"
    f = io.StringIO()
    with redirect_stdout(f):
        llvm_model.compile(fblocksize=blocksize)
    os.environ["LLEAVES_PRINT_UNOPTIMIZED_IR"] = "0"

    stdout = f.getvalue()
    # each cache block has an IR block called "instr-block-setup"
    assert "instr-block-setup:" in stdout
    if blocksize == 1:
        assert "instr-block-setup.1:" in stdout
        assert "instr-block-setup.99:" in stdout
        assert "instr-block-setup.100:" not in stdout
    if blocksize == 34:
        # NYC_taxi has 100 trees, hence blocksize 34 should create 3 blocks
        assert "instr-block-setup.1:" in stdout
        assert "instr-block-setup.2:" in stdout
        assert "instr-block-setup.3:" not in stdout
        assert "instr-block-setup.4:" not in stdout
    if blocksize == 100:
        assert "instr-block-setup.1:" not in stdout
        assert "instr-block-setup.2:" not in stdout

    np.testing.assert_almost_equal(
        llvm_model.predict(NYC_data[:1000], n_jobs=2),
        lgbm_model.predict(NYC_data[:1000], n_jobs=2),
    )


def test_small_codemodel(NYC_data):
    llvm_model = Model(model_file="tests/models/NYC_taxi/model.txt")
    lgbm_model = Booster(model_file="tests/models/NYC_taxi/model.txt")
    llvm_model.compile(fcodemodel="small")

    np.testing.assert_almost_equal(
        llvm_model.predict(NYC_data[:1000], n_jobs=2),
        lgbm_model.predict(NYC_data[:1000], n_jobs=2),
    )


def test_no_inline(NYC_data):
    llvm_model = Model(model_file="tests/models/NYC_taxi/model.txt")
    lgbm_model = Booster(model_file="tests/models/NYC_taxi/model.txt")
    llvm_model.compile(finline=False)

    np.testing.assert_almost_equal(
        llvm_model.predict(NYC_data[:1000], n_jobs=2),
        lgbm_model.predict(NYC_data[:1000], n_jobs=2),
    )



from pathlib import Path

import lightgbm as lgb
import numpy as np
import numpy.random
import pandas as pd
import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from lleaves import Model

numpy.random.seed(1337)


@pytest.fixture(scope="session", params=["pure_cat", "mixed_cat"])
def categorical_model_txt(tmpdir_factory, request):
    n_features = 5
    n_rows = 500
    # specify a tree and generate data from it
    if request.param == "pure_cat":
        n_categorical = 5

        def tree(axis):
            if axis[0] < 5:
                if axis[1] < 5:
                    return 2
                elif axis[0] < 2:
                    return 50
                else:
                    if axis[2] > 7:
                        return 10
                    return 2
            else:
                if axis[1] > 8:
                    return 99
                return 0

    else:
        n_categorical = 3

        def tree(axis):
            if axis[0] < 5:
                if axis[1] < 5:
                    return 2
                elif axis[1] < 2:
                    return 50
                else:
                    if axis[2] > 7:
                        return 100
                    return -32
            else:
                if axis[4] > 0.8:
                    return 99
                else:
                    if axis[3] < 0.2:
                        return -20
                return 0

    train_data_cat = np.random.randint(1, 10, size=(n_rows, n_categorical))
    if n_features - n_categorical > 0:
        train_data_num = np.random.rand(n_rows, n_features - n_categorical)
        train_data = np.concatenate([train_data_cat, train_data_num], axis=1)
    else:
        train_data = train_data_cat

    label = np.apply_along_axis(tree, axis=1, arr=train_data)
    train_data = lgb.Dataset(
        train_data, label=label, categorical_feature=list(range(n_categorical))
    )

    param = {}
    lightgbm_model = lgb.train(
        param, train_data, 1, categorical_feature=list(range(n_categorical))
    )

    tmpdir = tmpdir_factory.mktemp("model")
    model_path = tmpdir / "model.txt"
    lightgbm_model.save_model(str(model_path))
    return model_path


def test_large_categorical(tmpdir_factory):
    # test categorical var with >32 different entries
    def f(x):
        return x % 3

    train_data_cat = np.repeat(
        np.expand_dims(np.arange(1, 150), axis=1), repeats=40, axis=0
    )
    label = np.apply_along_axis(f, axis=1, arr=train_data_cat).flatten()
    train_data = lgb.Dataset(train_data_cat, label=label, categorical_feature=[0])
    lightgbm_model = lgb.train({}, train_data, 1, categorical_feature=[0])

    tmpdir = tmpdir_factory.mktemp("model")
    model_path = str(tmpdir / "model.txt")
    lightgbm_model.save_model(model_path)
    assert "decision_type=9 9 9" in Path(model_path).read_text()

    llvm_model = Model(model_file=model_path)
    llvm_model.compile()
    lgbm_model = lgb.Booster(model_file=model_path)
    tests_data = np.expand_dims(np.arange(0, 210), axis=1)
    numpy.testing.assert_equal(
        llvm_model.predict(tests_data), lgbm_model.predict(tests_data)
    )


def test_predict_pandas_categorical(tmpdir_factory):
    def f(x):
        t = ord(x["C1"]) - ord(x["C2"]) ** 2 + ord(x["C3"]) ** 3 + x["A"]
        return t / 100000

    train_df = pd.DataFrame(
        {
            "C1": 500 * ["a"] + 500 * ["b"] + 500 * ["c"],
            "C2": 500 * ["b"] + 500 * ["c"] + 500 * ["d"],
            "C3": 10 * (50 * ["x"] + 50 * ["y"] + 25 * ["z"] + 25 * ["w"]),
            "A": 5 * (100 * [10] + 100 * [-10] + 100 * [-30]),
        }
    )
    result = train_df.apply(f, axis=1)
    train_df["C1"] = train_df["C1"].astype("category")
    train_df["C2"] = train_df["C2"].astype("category")
    train_df["C3"] = train_df["C3"].astype("category")
    train_data = lgb.Dataset(train_df, label=result, categorical_feature="auto")
    lightgbm_model = lgb.train({}, train_data, 3, categorical_feature="auto")
    assert len(lightgbm_model.pandas_categorical) == 3

    tmpdir = tmpdir_factory.mktemp("model")
    model_path = str(tmpdir / "model.txt")
    lightgbm_model.save_model(model_path)

    llvm_model = Model(model_file=model_path)
    llvm_model.compile()
    lgbm_model = lgb.Booster(model_file=model_path)

    df = pd.DataFrame(
        {
            "C1": 2 * ["a", "b", "c", "c"],
            "C2": 2 * ["b", "c", "d", "b"],
            "C3": 2 * ["x", "y", "z", "w"],
            "A": [10, -10, 100, -30, -30, 10, 1, 100],
        },
    ).astype({"C1": "category", "C2": "category", "C3": "category"})
    df2 = df.copy()
    # reorder categories to ensure that same letter now maps to a different pandas code
    df2["C1"] = df2["C1"].cat.set_categories(["c", "b", "a"])
    df2["C2"] = df2["C2"].cat.set_categories(["a", "z", "w"])
    assert not (list(df["C1"].cat.codes) == list(df2["C1"].cat.codes))
    assert not (list(df["C2"].cat.codes) == list(df2["C2"].cat.codes))

    df3 = pd.DataFrame(
        {
            "C1": 2 * ["a", "b", "c", "s"],
            "C2": 2 * ["b", "c", "v", "b"],
            "C3": 2 * ["x", "u", "z", "w"],
            "A": [10, -10, 100, -30, -30, 10, 1, 100],
        },
    ).astype({"C1": "category", "C2": "category", "C3": "category"})

    # prediction from lightgbm should still be equal
    numpy.testing.assert_equal(lightgbm_model.predict(df), lightgbm_model.predict(df2))
    test_data = [
        df,
        df2,
        df3,
        train_df,
        np.array([[1, 1, 1, 10], [2, -2, 2, 10], [0, 1, 2, 10], [2, 1, 0, -10]]),
    ]
    for data in test_data:
        numpy.testing.assert_equal(llvm_model.predict(data), lgbm_model.predict(data))

    data = (
        pd.DataFrame(
            {
                "C1": 2 * ["b", "a", "a", "a"],
                "C2": 2 * ["d", "c", "a", "b"],
                "C3": 2 * ["z", "w", "z", "y"],
                "A": [10, -10, 100, -30, -30, 10, 1, 100],
            }
        ),
    )
    with pytest.raises(ValueError):
        # no categories
        llvm_model.predict(data)


@settings(deadline=1000)
@given(data=st.data())
def test_categorical_prediction_llvm_real(data, categorical_model_txt):
    lgbm_model = lgb.Booster(model_file=str(categorical_model_txt))
    llvm_model = Model(model_file=categorical_model_txt)
    llvm_model.compile()

    input = data.draw(
        st.lists(
            st.integers(min_value=0, max_value=2 ** 31 - 2),
            max_size=llvm_model.num_feature(),
            min_size=llvm_model.num_feature(),
        )
    )
    assert llvm_model.predict([input]) == lgbm_model.predict([input])


def test_pure_categorical_prediction():
    llvm_model = Model("tests/models/pure_categorical/model.txt")
    llvm_model.compile()
    lgbm_model = lgb.Booster(model_file="tests/models/pure_categorical/model.txt")

    results = [12.616231057968633, 10.048276920678525, 9.2489478721549396]
    for data, res_idx in zip(
        [
            [0, 9, 0],
            [0, -1, 0],
            [1, 9, 0],
            [0, 6, 5],
            [1, 5, 1],
            [2, 5, 1],
            [4, 5, 1],
            [5, 5, 9],
            [6, 5, 3],
            [9, 5, 2],
            [-1, 5, 2],
        ],
        [0, 1, 0, 0, 1, 1, 1, 2, 2, 2, 2],
    ):
        assert lgbm_model.predict([data]) == [results[res_idx]]
        assert llvm_model.predict([data]) == [results[res_idx]]

    na = float("NaN")
    inf = float("Inf")
    for data, res_idx in zip(
        [
            [na, 9.0, 0.0],
            [na, 0.0, 0.0],
            [inf, 0.0, 0.0],
            [0.0, na, 0.0],
            [4.0, inf, 0.0],
            [na, na, 0.0],
        ],
        [0, 2, 2, 1, 1, 2],
    ):
        assert lgbm_model.predict([data]) == [results[res_idx]]
        assert llvm_model.predict([data]) == [results[res_idx]]



import numpy as np
import pandas as pd

from benchmarks.benchmark import NYC_used_columns
from benchmarks.train_NYC_model import feature_enginering

df = pd.read_csv("airline_data_factorized.csv")
airline_X = df.to_numpy(np.float64)
np.save("airline.npy", airline_X)

df = pd.read_parquet("yellow_tripdata_2016-01.parquet", columns=NYC_used_columns)
NYC_X = feature_enginering().fit_transform(df).astype(np.float64)
np.save("NYC_taxi.npy", NYC_X)

df = pd.read_parquet("mtpl2.parquet")
mtpl2_X = df.to_numpy(np.float64)
np.save("mtpl2.npy", mtpl2_X)



import lightgbm as lgb
import numpy as np
import numpy.testing as npt
import pytest

import lleaves


# 2: MissingType None, default left
# 4: MissingType 0, default right
# 6: MissingType 0, default left
# 8: MissingType NaN, default left
# 10: MissingType NaN, default right
@pytest.mark.parametrize(
    "decision_type, threshold_le_zero",
    [
        (0, True),
        (2, True),
        (4, True),
        (6, True),
        (8, True),
        (10, True),
        (0, False),
        (2, False),
        (4, False),
        (6, False),
        (8, False),
        (10, False),
    ],
)
def test_zero_as_missing_numerical(tmp_path, decision_type, threshold_le_zero):
    model_txt = tmp_path / "model.txt"
    with open("tests/models/tiniest_single_tree/model.txt") as infile, open(
        model_txt, "w"
    ) as outfile:
        for line in infile.readlines():
            if line.startswith("decision_type="):
                outfile.write(line.replace("2", str(decision_type)))
            elif threshold_le_zero and line.startswith("threshold="):
                outfile.write(line.replace("0.", "-0."))
            else:
                outfile.write(line)

    lgbm_model = lgb.Booster(model_file=str(model_txt))
    llvm_model = lleaves.Model(model_file=str(model_txt))
    llvm_model.compile()

    nan = float("NaN")
    data = [
        [0.0, 1.0, 1.0],
        [0.0, -1.0, -1.0],
        [0.0, 1.0, 0.5],
        [0.0, -1.0, -0.5],
        [0.0, 0.5, 1.0],
        [0.0, -0.5, -1.0],
        [0.0, 0.5, 0.0],
        [0.0, -0.5, 0.0],
        [-0.01, -0.01, -0.01],
        [0.0, 0.0, 0.0],
        [0.01, 0.01, 0.01],
        [nan, nan, nan],
        [None, None, None],
    ]
    npt.assert_equal(llvm_model.predict(data), lgbm_model.predict(data))


@pytest.mark.parametrize(
    "decision_type, zero_in_bitvec",
    [
        (1, True),
        (3, True),
        (5, True),
        (7, True),
        (9, True),
        (11, True),
        (1, False),
        (3, False),
        (5, False),
        (7, False),
        (9, False),
        (11, False),
    ],
)
def test_zero_as_missing_categorical(tmp_path, decision_type, zero_in_bitvec):
    model_txt = tmp_path / "model.txt"
    with open("tests/models/pure_categorical/model.txt") as infile, open(
        model_txt, "w"
    ) as outfile:
        for line in infile.readlines():
            if line.startswith("decision_type"):
                outfile.write(line.replace("1", str(decision_type)))
            elif line.startswith("cat_threshold") and not zero_in_bitvec:
                outfile.write(line.replace("23", "22"))
            else:
                outfile.write(line)

    lgbm_model = lgb.Booster(model_file=str(model_txt))
    llvm_model = lleaves.Model(model_file=str(model_txt))
    llvm_model.compile()

    nan = float("NaN")
    data = [
        [1.0, 6.0, 0.0],
        [nan, 6.0, 0.0],
        [nan, 1.0, 0.0],
        [None, 1.0, 0.0],
        [1.0, nan, 0.0],
        [3.0, nan, 0.0],
        [3.0, None, 0.0],
        [0.0, 6.0, 0.0],
        [0.0, 1.0, 0.0],
        [1.0, 0.0, 0.0],
        [1.0, -0.001, 0.0],
        [1.0, 0.001, 0.0],
        [3.0, 0.0, 0.0],
        [3.0, -0.001, 0.0],
        [3.0, 0.001, 0.0],
        [nan, nan, nan],
        [None, None, None],
    ]
    npt.assert_equal(llvm_model.predict(data), lgbm_model.predict(data))


def test_lightgbm_nan_pred_inconsistency(tmp_path):
    # see https://github.com/dmlc/treelite/issues/277
    model_file = str(tmp_path / "model.txt")
    X = np.array(30 * [[1]] + 30 * [[2]] + 30 * [[0]])
    y = np.array(60 * [5] + 30 * [10])
    train_data = lgb.Dataset(X, label=y, categorical_feature=[0])
    bst = lgb.train({}, train_data, 1, categorical_feature=[0])
    bst.save_model(model_file)

    # just to make sure it's not due to LightGBM model export
    lgbm_model = lgb.Booster(model_file=model_file)
    llvm_model = lleaves.Model(model_file=model_file)
    llvm_model.compile()

    data = np.array([[np.NaN], [0.0], [-0.1], [0.1], [10.0], [np.Inf], [-np.NaN]])
    npt.assert_equal(lgbm_model.predict(data), llvm_model.predict(data))


def test_nan_prediction_numerical():
    model_path = "tests/models/tiniest_single_tree/model.txt"
    llvm_model = lleaves.Model(model_file=model_path)
    llvm_model.compile()
    lgbm_model = lgb.Booster(model_file=model_path)

    nan = float("NaN")
    inf = float("Inf")
    data = [
        3 * [nan],
        3 * [inf],
        [nan, inf, nan],
        [0.0, nan, 0.0],
        [0.0, inf, 0.0],
        [0.0, nan, 1.0],
        [0.0, inf, 1.0],
        [0.0, 0.0, nan],
        [0.0, 0.0, inf],
        [0.0, 1.0, nan],
        [0.0, 1.0, inf],
    ]
    npt.assert_equal(llvm_model.predict(data), lgbm_model.predict(data))


def test_nan_prediction_categorical():
    model_path = "tests/models/pure_categorical/model.txt"
    llvm_model = lleaves.Model(model_file=model_path)
    llvm_model.compile()
    lgbm_model = lgb.Booster(model_file=model_path)

    nan = float("NaN")
    inf = float("Inf")
    data = [
        3 * [nan],
        3 * [inf],
        [nan, inf, nan],
        # run both branches with Nans
        [1.0, nan, 0.0],
        [1.0, inf, 0.0],
        [0.0, nan, 0.0],
        [0.0, inf, 0.0],
        [nan, 6.0, 0.0],
        [inf, 6.0, 0.0],
        [nan, 0.0, 0.0],
        [inf, 0.0, 0.0],
    ]
    npt.assert_equal(llvm_model.predict(data), lgbm_model.predict(data))



import os
import time

import lleaves

model = os.environ["LLEAVES_BENCHMARK_MODEL"]

fcodemodel = os.environ.get("LLEAVES_FCODEMODEL", "large")
finline = os.environ.get("LLEAVES_FINLINE", "True")
assert finline in (None, "True", "False")
fblocksize = os.environ.get("LLEAVES_FBLOCKSIZE", 34)

print(f"Generating {model}.o")

llvm_model = lleaves.Model(
    model_file=f"../../../tests/models/{model}/model.txt",
)
start = time.time()
llvm_model.compile(
    cache=f"../{model}.o",
    fblocksize=int(fblocksize) if fblocksize else None,
    fcodemodel=fcodemodel,
    finline=finline == "True",
)
print(f"Compiling took: {time.time() - start}")
